# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# DO NOT IMPORT GTK HERE: see
#  http://lists.partiwm.org/pipermail/parti-discuss/2008-September/000041.html
#  http://lists.partiwm.org/pipermail/parti-discuss/2008-September/000042.html
# (also do not import anything that imports gtk)
import glob
import shlex
import signal
from time import monotonic
from typing import NoReturn
from subprocess import Popen, PIPE, call
import os.path

from xpra.common import RESOLUTION_ALIASES, DEFAULT_REFRESH_RATE, get_refresh_rate_for_value
from xpra.scripts.config import InitException, get_Xdummy_confdir, FALSE_OPTIONS
from xpra.util.str_fn import csv
from xpra.util.env import envint, envbool, shellsub, osexpand, get_exec_env, get_saved_env_var
from xpra.os_util import getuid, getgid, POSIX, OSX
from xpra.server.util import setuidgid
from xpra.util.io import is_writable, pollwait
from xpra.platform.displayfd import read_displayfd, parse_displayfd
from xpra.log import Logger


VFB_WAIT = envint("XPRA_VFB_WAIT", 3)
XVFB_EXTRA_ARGS = os.environ.get("XPRA_XVFB_EXTRA_ARGS", "")


def parse_resolution(res_str, default_refresh_rate=DEFAULT_REFRESH_RATE//1000) -> tuple[int, int, int] | None:
    if not res_str:
        return None
    s = res_str.upper()       # ie: 4K60
    res_part = s
    hz = get_refresh_rate_for_value(str(default_refresh_rate), DEFAULT_REFRESH_RATE)//1000
    for sep in ("@", "K", "P"):
        pos = s.find(sep)
        if 0 < pos < len(s)-1:
            res_part, hz = s.split(sep, 1)
            if sep != "@":
                res_part += sep
            break
    if res_part in RESOLUTION_ALIASES:
        w, h = RESOLUTION_ALIASES[res_part]
    else:
        try:
            parts = tuple(int(x) for x in res_part.replace(",", "x").split("X", 1))
        except ValueError:
            raise ValueError(f"failed to parse resolution {res_str!r}") from None
        if len(parts) != 2:
            raise ValueError(f"invalid resolution string {res_str!r}")
        w = parts[0]
        h = parts[1]
    return w, h, int(hz)


def parse_resolutions(s, default_refresh_rate=DEFAULT_REFRESH_RATE//1000) -> tuple | None:
    if not s or s.lower() in FALSE_OPTIONS:
        return None
    if s.lower() in ("none", "default"):
        return ()
    return tuple(parse_resolution(v, default_refresh_rate) for v in s.split(","))


def parse_env_resolutions(envkey="XPRA_DEFAULT_VFB_RESOLUTIONS",
                          single_envkey="XPRA_DEFAULT_VFB_RESOLUTION",
                          default_res="8192x4096",
                          default_refresh_rate=DEFAULT_REFRESH_RATE//1000):
    s = os.environ.get(envkey)
    if s:
        return parse_resolutions(s, default_refresh_rate)
    return (parse_resolution(os.environ.get(single_envkey, default_res), default_refresh_rate), )


def get_desktop_vfb_resolutions(default_refresh_rate=DEFAULT_REFRESH_RATE//1000):
    return parse_env_resolutions("XPRA_DEFAULT_DESKTOP_VFB_RESOLUTIONS",
                                 "XPRA_DEFAULT_DESKTOP_VFB_RESOLUTION",
                                 "1280x1024",
                                 default_refresh_rate=default_refresh_rate)


PRIVATE_XAUTH = envbool("XPRA_PRIVATE_XAUTH", False)
XAUTH_PER_DISPLAY = envbool("XPRA_XAUTH_PER_DISPLAY", True)


vfb_logger = None


def get_vfb_logger() -> Logger:
    global vfb_logger
    if not vfb_logger:
        vfb_logger = Logger("server", "x11", "screen")
    return vfb_logger


def osclose(fd: int) -> None:
    try:
        os.close(fd)
    except OSError:
        pass


def create_xorg_device_configs(xorg_conf_dir: str, device_uuid: str, uid: int, gid: int) -> None:
    log = get_vfb_logger()
    log("create_xorg_device_configs(%s, %s, %i, %i)", xorg_conf_dir, device_uuid, uid, gid)
    if not device_uuid:
        return

    def makedir(dirname) -> None:
        log("makedir(%s)", dirname)
        os.mkdir(dirname)
        os.lchown(dirname, uid, gid)

    # create conf dir if needed:
    d = xorg_conf_dir
    dirs: list[str] = []
    while d and not os.path.exists(d):
        log("create_device_configs: dir does not exist: %s", d)
        dirs.insert(0, d)
        d = os.path.dirname(d)
    for d in dirs:
        makedir(d)

    conf_files: list[str] = []
    for i, dev_type in (
        (0, "pointer"),
        (1, "touchpad"),
    ):
        f = save_input_conf(xorg_conf_dir, i, dev_type, device_uuid, uid, gid)
        conf_files.append(f)


def save_input_conf(xorg_conf_dir: str, i, dev_type, device_uuid: str, uid: int, gid: int) -> str:
    # create individual device files:
    upper_dev_type = dev_type[:1].upper()+dev_type[1:]   # ie: Pointer
    product_name = f"Xpra Virtual {upper_dev_type} {device_uuid}"
    identifier = f"xpra-virtual-{dev_type}"
    conf_file = os.path.join(xorg_conf_dir, f"{i:02}-{dev_type}.conf")
    with open(conf_file, "w", encoding="utf8") as f:
        f.write(f"""Section "InputClass"
Identifier "{identifier}"
MatchProduct "{product_name}"
MatchUSBID "ffff:ffff"
MatchIs{upper_dev_type} "True"
Driver "libinput"
Option "AccelProfile" "flat"
Option "Ignore" "False"
EndSection
""")
        os.fchown(f.fileno(), uid, gid)
    # Option "AccelerationProfile" "-1"
    # Option "AccelerationScheme" "none"
    # Option "AccelSpeed" "-1"
    return conf_file


def valid_xauth(filename: str, uid: int = getuid(), gid: int = getgid()) -> str:
    if not filename:
        return ""
    if not os.path.exists(filename):
        return ""
    if not is_writable(filename, uid, gid):
        log = get_vfb_logger()
        log.info(f"ignoring non-writable XAUTHORITY={filename!r}")
        return ""
    return filename


def get_xauthority_path(display_name: str) -> str:
    assert POSIX
    # pylint: disable=import-outside-toplevel
    from xpra.platform.posix.paths import _get_xpra_runtime_dir
    expanded_home = os.path.expanduser("~")
    has_home = os.path.exists(expanded_home) and is_writable(expanded_home, getuid(), getgid())
    if PRIVATE_XAUTH or (not has_home and os.environ.get("XDG_RUNTIME_DIR")):
        d = _get_xpra_runtime_dir()
        if XAUTH_PER_DISPLAY:
            filename = "Xauthority-" + display_name.lstrip(":")
        else:
            filename = "Xauthority"
    else:
        if has_home and is_writable(expanded_home, getuid(), getgid()):
            d = expanded_home
        else:
            d = os.environ.get("TMPDIR", "/tmp")
        filename = ".Xauthority"
    return os.path.join(d, filename)


def get_xvfb_env(xvfb_executable: str) -> dict[str, str]:
    keep = tuple(f"^{envname}$" for envname in (
        "SHELL", "HOSTNAME", "XMODIFIERS",
        "PWD", "HOME", "USERNAME", "LANG", "TERM", "USER",
        "XDG_RUNTIME_DIR", "XDG_DATA_DIR",
        "PATH", "LD_LIBRARY_PATH",
        "XAUTHORITY",
        "XPRA_SESSION_DIR",
    ))
    env = get_exec_env(keep=keep)
    if xvfb_executable.endswith("Xephyr"):
        env["DISPLAY"] = get_saved_env_var("DISPLAY")
    return env


def patch_xvfb_command(xvfb_cmd: list[str], w: int, h: int, pixel_depth: int):
    # find the ["-screen"] or ["screen", "0"] arguments:
    try:
        screen_arg = xvfb_cmd.index("-screen")
    except ValueError:
        # not found!
        screen_arg = -1
    if screen_arg > 0:
        # remove the screen number if specified:
        try:
            no_arg = xvfb_cmd.index("0")
        except ValueError:
            no_arg = -1
        if no_arg > 0 and no_arg == screen_arg + 1:
            xvfb_cmd.pop(no_arg)
        # remove the "-screen" argument and the geometry that follows:
        xvfb_cmd.pop(screen_arg)
        xvfb_cmd.pop(screen_arg)
    # this is the geometry we want to add:
    if xvfb_cmd[0].endswith("Xvfb"):
        # ie: "-screen 0 8192x4096x32"
        xvfb_cmd += ["-screen", "0", f"{w}x{h}x{pixel_depth}"]
    else:
        # play it safe, don't specify the screen number:
        xvfb_cmd += ["-screen", f"{w}x{h}x{pixel_depth}"]


def start_Xvfb(xvfb_cmd: list[str], vfb_geom, pixel_depth: int, display_name: str, cwd,
               uid: int, gid: int, username: str, uinput_uuid="") -> tuple[Popen, str]:
    if not POSIX:
        raise InitException(f"starting an Xvfb is not supported on {os.name}")
    if OSX:
        raise InitException("starting an Xvfb is not supported on MacOS")
    if not xvfb_cmd:
        raise InitException("the 'xvfb' command is not defined")
    if not display_name:
        raise ValueError("no display name")

    log = get_vfb_logger()
    log("start_Xvfb%s XVFB_EXTRA_ARGS=%s",
        (xvfb_cmd, vfb_geom, pixel_depth, display_name, cwd, uid, gid, username, uinput_uuid),
        XVFB_EXTRA_ARGS)
    use_display_fd = display_name[0] == "S"
    if XVFB_EXTRA_ARGS:
        xvfb_cmd += shlex.split(XVFB_EXTRA_ARGS)

    subs: dict[str, str] = {}

    def pathexpand(s: str) -> str:
        return osexpand(s, actual_username=username, uid=uid, gid=gid, subs=subs)
    etc_prefix = os.environ.get("XPRA_INSTALL_PREFIX", "")
    if etc_prefix.endswith("/usr"):
        etc_prefix = etc_prefix[:-4]
    subs |= {
        "DISPLAY": display_name,
        "XPRA_LOG_DIR": pathexpand(os.environ.get("XPRA_LOG_DIR", "")),
        "XORG_CONFIG_PREFIX": os.environ.get("XORG_CONFIG_PREFIX", etc_prefix),
    }

    # identify logfile argument if it exists,
    # as we may have to rename it, or create the directory for it:
    # make sure all path values are expanded:
    xvfb_cmd = [pathexpand(s) for s in xvfb_cmd]

    # try to honour initial geometries if specified:
    xvfb_executable = xvfb_cmd[0]
    if len(vfb_geom) >= 2 and (xvfb_executable.endswith("Xvfb") or xvfb_executable.endswith("Xephyr")):
        w, h = vfb_geom[:2]
        log("patch_xvfb_command%s", (xvfb_cmd, w, h, pixel_depth or 32))
        patch_xvfb_command(xvfb_cmd, w, h, pixel_depth or 32)
        log(f"{xvfb_cmd=!r}")
    try:
        logfile_argindex = xvfb_cmd.index('-logfile')
        if logfile_argindex+1 >= len(xvfb_cmd):
            raise InitException("invalid xvfb command string: -logfile should not be last")
        xorg_log_file = xvfb_cmd[logfile_argindex+1]
    except ValueError:
        xorg_log_file = None
    tmp_xorg_log_file = None
    if xorg_log_file:
        if use_display_fd:
            # keep track of it, so that we can rename it later:
            tmp_xorg_log_file = xorg_log_file
        # make sure the Xorg log directory exists:
        xorg_log_dir = os.path.dirname(xorg_log_file)
        if not os.path.exists(xorg_log_dir):
            try:
                log("creating Xorg log dir '%s'", xorg_log_dir)
                os.mkdir(xorg_log_dir, 0o700)
                if POSIX and uid != getuid() or gid != getgid():
                    try:
                        os.lchown(xorg_log_dir, uid, gid)
                    except OSError:
                        log("lchown(%s, %i, %i)", xorg_log_dir, uid, gid, exc_info=True)
            except OSError as e:
                raise InitException(f"failed to create the Xorg log directory {xorg_log_dir!r}: {e}") from None

    if uinput_uuid:
        # use uinput:
        # identify -config xorg.conf argument and replace it with the uinput one:
        try:
            config_argindex = xvfb_cmd.index("-config")
        except ValueError:
            log.warn("Warning: cannot use uinput")
            log.warn(" '-config' argument not found in the xvfb command")
        else:
            if config_argindex+1 >= len(xvfb_cmd):
                raise InitException("invalid xvfb command string: -config should not be last")
            xorg_conf = xvfb_cmd[config_argindex+1]
            if xorg_conf.endswith("xorg.conf"):
                xorg_conf = xorg_conf.replace("xorg.conf", "xorg-uinput.conf")
                if os.path.exists(xorg_conf):
                    xvfb_cmd[config_argindex+1] = xorg_conf
            # create uinput device definition files:
            # (we have to assume that Xorg is configured to use this path..)
            xorg_conf_dir = pathexpand(get_Xdummy_confdir())
            create_xorg_device_configs(xorg_conf_dir, uinput_uuid, uid, gid)

    if (xvfb_executable.endswith("Xorg") or xvfb_executable.endswith("Xdummy")) and pixel_depth > 0:
        xvfb_cmd.append("-depth")
        xvfb_cmd.append(str(pixel_depth))
    env = get_xvfb_env(xvfb_executable)
    log(f"xvfb env={env}")
    xvfb = None
    try:
        if use_display_fd:
            def displayfd_err(msg: str) -> NoReturn:
                if xvfb and xvfb.poll() is None:
                    log.error(" stopping vfb process with pid %i", xvfb.pid)
                    xvfb.terminate()
                raise InitException(f"{xvfb_executable}: {msg}")
            r_pipe, w_pipe = os.pipe()
            try:
                os.set_inheritable(w_pipe, True)
                xvfb_cmd += ["-displayfd", str(w_pipe)]
                xvfb_cmd[0] = f"{xvfb_executable}-for-Xpra-" + display_name.lstrip(":")

                def preexec() -> None:
                    os.setpgrp()
                    if getuid() == 0 and uid:
                        setuidgid(uid, gid)
                try:
                    # pylint: disable=consider-using-with
                    # pylint: disable=subprocess-popen-preexec-fn
                    xvfb = Popen(xvfb_cmd, executable=xvfb_executable,
                                 preexec_fn=preexec, cwd=cwd, env=env, pass_fds=(w_pipe,))
                except OSError as e:
                    log("Popen%s", (xvfb_cmd, xvfb_executable, cwd), exc_info=True)
                    raise InitException(f"failed to execute xvfb command {xvfb_cmd}: {e}") from None
                if xvfb.poll() is not None:
                    raise InitException(f"xvfb command has terminated: {xvfb_cmd}")
                # Read the display number from the pipe we gave to Xvfb
                try:
                    buf = read_displayfd(r_pipe, proc=xvfb)
                except Exception as e:
                    buf = b""
                    log("read_displayfd(%s)", r_pipe, exc_info=True)
                    displayfd_err(f"failed to read displayfd pipe {r_pipe}: {e}")
            finally:
                osclose(r_pipe)
                osclose(w_pipe)
            n = parse_displayfd(buf, displayfd_err)
            if n < 0:
                displayfd_err(f"failed to parse displayfd output {buf!r}")
            new_display_name = f":{n}"
            log(f"Using display number provided by {xvfb_executable}: {new_display_name}")
            if tmp_xorg_log_file:
                # ie: ${HOME}/.xpra/Xorg.${DISPLAY}.log -> /home/antoine/.xpra/Xorg.S14700.log
                f0 = shellsub(tmp_xorg_log_file, subs)
                subs["DISPLAY"] = new_display_name
                # ie: ${HOME}/.xpra/Xorg.${DISPLAY}.log -> /home/antoine/.xpra/Xorg.:1.log
                f1 = shellsub(tmp_xorg_log_file, subs)
                if f0 != f1:
                    try:
                        os.rename(f0, f1)
                    except Exception as e:
                        log.warn("Warning: failed to rename Xorg log file,")
                        log.warn(f" from {f0!r} to {f1!r}")
                        log.warn(f" {e}")
            display_name = new_display_name
        else:
            # use display specified
            xvfb_cmd[0] = f"{xvfb_executable}-for-Xpra-"+display_name.lstrip(":")
            xvfb_cmd.append(display_name)

            def preexec() -> None:
                if getuid() == 0 and (uid != 0 or gid != 0):
                    setuidgid(uid, gid)
                else:
                    os.setsid()
            log("xvfb_cmd=%s", xvfb_cmd)
            # pylint: disable=consider-using-with
            # pylint: disable=subprocess-popen-preexec-fn
            xvfb = Popen(xvfb_cmd, executable=xvfb_executable,
                         stdin=PIPE, preexec_fn=preexec, env=env)
    except Exception:
        if xvfb and xvfb.poll() is None:
            log.error(" stopping vfb process with pid %i", xvfb.pid)
            xvfb.terminate()
        raise
    log("xvfb process=%s", xvfb)
    log("display_name=%s", display_name)
    return xvfb, display_name


def start_xvfb_standalone(xvfb_cmd: list[str], sessions_dir: str, pixel_depth=0, display_name="") -> int:
    # we may have to tweak the environment to emulate having an xpra session
    # as the default xvfb commands may refer to $XAUTHORITY and $XPRA_SESSION_DIR
    xauthority = os.environ.get("XAUTHORITY", "")
    if "$XAUTHORITY" in xvfb_cmd and not valid_xauth(xauthority):
        xauthority = osexpand(get_xauthority_path(display_name))
        os.environ["XAUTHORITY"] = xauthority
    if "$XPRA_SESSION_DIR" in xvfb_cmd and "XPRA_SESSION_DIR" not in os.environ:
        from xpra.scripts.session import make_session_dir
        session_dir = make_session_dir("xvfb", sessions_dir, display_name)
        os.environ["XPRA_SESSION_DIR"] = session_dir
    vfb_geom = ()
    cwd = os.getcwd()
    xvfb, actual_display_name = start_Xvfb(xvfb_cmd, vfb_geom, pixel_depth, display_name,
                                           cwd, uid=getuid(), gid=getgid(), username="", uinput_uuid="")
    if actual_display_name != display_name:
        print(f"display {actual_display_name!r} started")
    print(f"xvfb pid: {xvfb.pid}")
    return xvfb.wait()


def kill_xvfb(xvfb_pid: int) -> None:
    log = get_vfb_logger()
    log.info("killing xvfb with pid %s", xvfb_pid)
    try:
        os.kill(xvfb_pid, signal.SIGTERM)
    except OSError as e:
        log.info("failed to kill xvfb process with pid %s:", xvfb_pid)
        log.info(" %s", e)
    xauthority = os.environ.get("XAUTHORITY", "")
    if PRIVATE_XAUTH and xauthority and os.path.exists(xauthority):
        os.unlink(xauthority)


def set_initial_resolution(resolutions, dpi: int = 0) -> None:
    log = get_vfb_logger()
    log("set_initial_resolution(%s)", resolutions)
    try:
        # pylint: disable=import-outside-toplevel
        from xpra.x11.bindings.randr import RandRBindings
        # try to set a reasonable display size:
        randr = RandRBindings()
        if not randr.has_randr():
            log.warn("Warning: no RandR support,")
            log.warn(" default virtual display size unchanged")
            return
        if randr.is_dummy16():
            monitors = {}
            x, y = 0, 0
            for i, res in enumerate(resolutions):
                assert len(res) == 3
                if not all(isinstance(v, int) for v in res):
                    raise ValueError(f"resolution values must be ints, found: {res} ({csv(type(v) for v in res)})")
                w, h, hz = res
                mdpi = dpi
                # guess the DPI if we don't have one:
                if mdpi <= 0:
                    # use a higher DPI for higher resolutions
                    if w >= 4096 or h >= 2560:
                        mdpi = 144
                    elif w >= 2560 or h >= 1440:
                        mdpi = 120
                    else:
                        mdpi = 96

                def rdpi(v: int) -> int:
                    return round(v * 25.4 / mdpi)
                monitors[i] = {
                    "name": f"VFB-{i}",
                    "primary": i == 0,
                    "geometry": (x, y, w, h),
                    "width-mm": rdpi(w),
                    "height-mm": rdpi(h),
                    "refresh-rate": hz*1000,
                    "automatic": True,
                }
                x += w
                # arranging vertically:
                # y += h
            randr.set_crtc_config(monitors)
            return
        res = resolutions[0][:2]
        sizes = randr.get_xrr_screen_sizes()
        size = randr.get_screen_size()
        log(f"RandR available, current {size=}, sizes available={sizes}")
        if res not in sizes:
            log.warn(f"Warning: cannot set resolution to {res}")
            log.warn(" (this resolution is not available)")
        elif res == size:
            log(f"initial resolution already set: {res}")
        else:
            log(f"RandR setting new screen size to {res}")
            randr.set_screen_size(*res)
    except Exception as e:
        log(f"set_initial_resolution({resolutions})", exc_info=True)
        log.error("Error: failed to set the default screen size:")
        log.estr(e)


def xauth_add(filename: str, display_name: str, xauth_data: str, uid: int, gid: int) -> None:
    xauth_args = ["-f", filename, "add", display_name, "MIT-MAGIC-COOKIE-1", xauth_data]
    xauth_cmd = ["xauth"] + xauth_args
    try:
        def preexec():
            os.setsid()
            if getuid() == 0 and uid:
                setuidgid(uid, gid)
        start = monotonic()
        log = get_vfb_logger()
        log("xauth command: %s", xauth_cmd)
        code = call(xauth_cmd, preexec_fn=preexec)
        end = monotonic()
        elapsed = round(end-start)
        if code != 0 and elapsed >= 10:
            log.warn(f"Warning: xauth command took {elapsed} seconds and failed")
            # took more than 10 seconds to fail, check for stale locks:
            if glob.glob(f"{filename}-*"):
                log.warn("Warning: trying to clean some stale xauth locks")
                xauth_cmd = ["xauth", "-b"]+xauth_args
                log(f"xauth command: {xauth_cmd}")
                code = call(xauth_cmd, preexec_fn=preexec)
        if code != 0:
            raise OSError(f"non-zero exit code: {code}")
    except OSError as e:
        # trying to continue anyway!
        log = get_vfb_logger()
        log(f"xauth_add%s xauth_cmd={xauth_cmd}", (filename, display_name, xauth_data, uid, gid))
        log.error(f"Error adding xauth entry for {display_name}")
        log.error(" using command \"%s\":" % (" ".join(xauth_cmd)))
        log.estr(e)


def check_xvfb_process(xvfb=None, cmd: str = "Xvfb", timeout: int = 0, command=()) -> bool:
    if xvfb is None:
        # we don't have a process to check
        return True
    if pollwait(xvfb, timeout) is None:
        # process is running
        return True
    log = get_vfb_logger()
    log.error("")
    log.error("%s command has terminated! xpra cannot continue", cmd)
    log.error(" if the display is already running, try a different one,")
    log.error(" if the `xvfb` command is invalid, try a different one")
    if command:
        log.error(" full command: %r", shlex.join(command))
    log.error("")
    return False


def verify_display_ready(xvfb, display_name: str, shadowing_check=True, log_errors=True, timeout=VFB_WAIT) -> bool:
    from xpra.x11.bindings.wait_for_x_server import wait_for_x_server  # pylint: disable=import-outside-toplevel
    # Whether we spawned our server or not, it is now running -- or at least
    # starting.  First wait for it to start up:
    try:
        wait_for_x_server(display_name, timeout)
    except Exception as e:
        log = get_vfb_logger()
        log("verify_display_ready%s", (xvfb, display_name, shadowing_check), exc_info=True)
        if log_errors:
            log.error(f"Error: failed to connect to display {display_name!r}")
            log.estr(e)
        return False
    if shadowing_check and not check_xvfb_process(xvfb):
        # if we're here, there is an X11 server, but it isn't the one we started!
        log = get_vfb_logger()
        log("verify_display_ready%s display exists, but the vfb process has terminated",
            (xvfb, display_name, shadowing_check, log_errors))
        if log_errors:
            log.error(f"There is an X11 server already running on display {display_name}:")
            log.error("You may want to use:")
            log.error(f"  'xpra upgrade {display_name}' if an instance of xpra is still connected to it")
            log.error(f"  'xpra --use-display start {display_name}' to connect xpra to an existing X11 server only")
            log.error("")
        return False
    return True


def main() -> None:
    # pylint: disable=import-outside-toplevel
    import sys
    display = ""
    if len(sys.argv) > 1:
        display = sys.argv[1]
    from xpra.x11.bindings.wait_for_x_server import wait_for_x_server
    wait_for_x_server(display, VFB_WAIT)
    print("OK")


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import subprocess
from queue import SimpleQueue, Empty
from collections.abc import Callable
from typing import Any

from xpra.gtk.signals import register_os_signals
from xpra.util.str_fn import csv, repr_ellipsized, hexstr
from xpra.util.env import envint, envbool, get_exec_env
from xpra.net.bytestreams import TwoFileConnection
from xpra.net.common import ConnectionClosedException, PacketType, PacketElement
from xpra.net.protocol.socket_handler import SocketProtocol
from xpra.net.protocol.constants import CONNECTION_LOST, GIBBERISH
from xpra.common import noop
from xpra.os_util import gi_import, WIN32
from xpra.util.io import setbinarymode
from xpra.util.system import SIGNAMES
from xpra.util.child_reaper import getChildReaper
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("util")

# this wrapper allows us to interact with a subprocess as if it was
# a normal class with gobject signals
# so that we can interact with it using a standard xpra protocol layer
# there is a wrapper for the caller
# and one for the class
# they talk to each other through stdin / stdout,
# using the protocol for encoding the data


DEBUG_WRAPPER = envbool("XPRA_WRAPPER_DEBUG", False)
# to make it possible to inspect files (more human-readable):
HEXLIFY_PACKETS = envbool("XPRA_HEXLIFY_PACKETS", False)
# avoids showing a new console window on win32:
WIN32_SHOWWINDOW = envbool("XPRA_WIN32_SHOWWINDOW", False)

FLUSH = envbool("XPRA_SUBPROCESS_FLUSH", False)

FAULT_RATE = envint("XPRA_WRAPPER_FAULT_INJECTION_RATE")


INJECT_FAULT = noop
if FAULT_RATE > 0:
    _counter = 0

    def do_inject_fault(p) -> None:
        global _counter
        _counter += 1
        if (_counter % FAULT_RATE) == 0:
            log.warn("injecting fault in %s", p)
            p.raw_write("junk", "Wrapper JUNK! added by fault injection code")

    INJECT_FAULT = do_inject_fault


def setup_fastencoder_nocompression(protocol) -> None:
    from xpra.net.packet_encoding import get_enabled_encoders, PERFORMANCE_ORDER
    encoders = get_enabled_encoders(PERFORMANCE_ORDER)
    assert len(encoders) > 0, "no packet encoders available!?"
    for encoder in encoders:
        try:
            protocol.enable_encoder(encoder)
            log("protocol using %s", encoder)
            break
        except Exception as e:
            log("failed to enable %s: %s", encoder, e)
    # we assume this is local, so no compression:
    protocol.enable_compressor("none")


class SubprocessCallee:
    """
    This is the callee side, wrapping the gobject we want to interact with.
    All the input received will be converted to method calls on the wrapped object.
    Subclasses should register the signal handlers they want to see exported back to the caller.
    The convenience connect_export(signal-name, *args) can be used to forward signals unmodified.
    You can also call send() to pass packets back to the caller.
    (there is no validation of which signals are valid or not)
    """

    def __init__(self, input_filename="-", output_filename="-", wrapped_object=None, method_whitelist=None):
        self.name = ""
        self._input = None
        self._output = None
        self.input_filename = input_filename
        self.output_filename = output_filename
        self.method_whitelist = method_whitelist
        self.large_packets = []
        # the gobject instance which is wrapped:
        self.wrapped_object = wrapped_object
        self.send_queue = SimpleQueue()
        self.protocol = None
        register_os_signals(self.handle_signal, self.name)
        self.setup_mainloop()

    def setup_mainloop(self) -> None:
        self.mainloop = GLib.MainLoop()

    def connect_export(self, signal_name: str, *user_data) -> None:
        """ gobject style signal registration for the wrapped object,
            the signals will automatically be forwarded to the wrapper process
            using send(signal_name, *signal_args, *user_data)
        """
        log("connect_export%s", [signal_name] + list(user_data))
        args = list(user_data) + [signal_name]
        self.wrapped_object.connect(signal_name, self.export, *args)

    def export(self, *args) -> None:
        signal_name = args[-1]
        log("export(%s, ...)", signal_name)
        data = args[1:-1]
        self.send(signal_name, *tuple(data))

    def start(self) -> int:
        self.protocol = self.make_protocol()
        self.protocol.start()
        try:
            self.run()
            return 0
        except KeyboardInterrupt as e:
            log("start() KeyboardInterrupt %s", e)
            if str(e):
                log.warn("%s", e)
            return 0
        except RuntimeError:
            log.error("Error in subprocess main loop", exc_info=True)
            return 1
        finally:
            log("run() ended, calling cleanup and protocol close")
            self.cleanup()
            if self.protocol:
                self.protocol.close()
                self.protocol = None
            i = self._input
            if i:
                self._input = None
                try:
                    i.close()
                except OSError:
                    log("%s.close()", i, exc_info=True)
            o = self._output
            if o:
                self._output = None
                try:
                    o.close()
                except OSError:
                    log("%s.close()", o, exc_info=True)

    def make_protocol(self) -> SocketProtocol:
        # figure out where we read from and write to:
        if self.input_filename == "-":
            # disable stdin buffering:
            self._input = os.fdopen(sys.stdin.fileno(), "rb", 0)
            setbinarymode(self._input.fileno())
        else:
            self._input = open(self.input_filename, "rb")
        if self.output_filename == "-":
            # disable stdout buffering:
            self._output = os.fdopen(sys.stdout.fileno(), "wb", 0)
            setbinarymode(self._output.fileno())
        else:
            self._output = open(self.output_filename, "wb")
        # stdin and stdout wrapper:
        conn = TwoFileConnection(self._output, self._input,
                                 abort_test=None, target=self.name,
                                 socktype=self.name, close_cb=self.net_stop)
        conn.timeout = 0
        protocol = SocketProtocol(conn, self.process_packet, get_packet_cb=self.get_packet)
        setup_fastencoder_nocompression(protocol)
        protocol.large_packets = self.large_packets
        return protocol

    def run(self) -> None:
        self.mainloop.run()

    def net_stop(self) -> None:
        # this is called from the network thread,
        # we use idle add to ensure we clean things up from the main thread
        log("net_stop() will call stop from main thread")
        GLib.idle_add(self.stop)

    def cleanup(self) -> None:
        """ subclasses may override this method """

    def stop(self) -> None:
        p = self.protocol
        log("stop() protocol=%s", p)
        self.cleanup()
        if p:
            self.protocol = None
            p.close()
        self.do_stop()

    def do_stop(self) -> None:
        log("do_stop() stopping mainloop %s", self.mainloop)
        self.mainloop.quit()

    def handle_signal(self, sig) -> None:
        """ This is for OS signals SIGINT and SIGTERM """
        # next time, just stop:
        register_os_signals(self.signal_stop, self.name)
        signame = SIGNAMES.get(sig, sig)
        log("handle_signal(%s) calling stop from main thread", signame)
        self.send("signal", signame)
        GLib.timeout_add(0, self.cleanup)
        # give time for the network layer to send the signal message
        GLib.timeout_add(150, self.stop)

    def signal_stop(self, sig) -> None:
        """ This time we really want to exit without waiting """
        signame = SIGNAMES.get(sig, sig)
        log("signal_stop(%s) calling stop", signame)
        self.stop()

    def send(self, packet_type: str, *args: PacketElement) -> None:
        if HEXLIFY_PACKETS:
            args = [packet_type] + [hexstr(str(x)[:32]) for x in args]
        log("send: adding '%s' message (%s items already in queue)", packet_type, self.send_queue.qsize())
        packet = (packet_type, *args)
        self.send_queue.put(packet)
        p = self.protocol
        if p:
            p.source_has_more()
        INJECT_FAULT(p)

    def get_packet(self) -> tuple[PacketType, bool, bool]:
        try:
            item = self.send_queue.get(False)
            more = self.send_queue.qsize() > 0
        except Empty:
            item = None
            more = False
        return item, False, more

    def process_packet(self, proto, packet: PacketType) -> None:
        command = str(packet[0])
        if command == CONNECTION_LOST:
            log("connection-lost: %s, calling stop", packet[1:])
            self.net_stop()
            return
        if command == GIBBERISH:
            log.warn("Warning: gibberish received:")
            log.warn(" %s", repr_ellipsized(packet[1], limit=80))
            log.warn(" stopping")
            self.net_stop()
            return
        if command == "stop":
            log("received stop message")
            self.net_stop()
            return
        if command == "exit":
            log("received exit message")
            sys.exit(0)
        # make it easier to hookup signals to methods:
        attr = command.replace("-", "_")
        if self.method_whitelist is not None and attr not in self.method_whitelist:
            log.warn("invalid command %r, not in whitelist: %s", attr, csv(self.method_whitelist))
            return
        wo = self.wrapped_object
        if not wo:
            log("wrapped object is no more, ignoring method call '%s'", attr)
            return
        method = getattr(wo, attr, None)
        if not method:
            log.warn("unknown command: '%s'", attr)
            log.warn(" packet: '%s'", repr_ellipsized(str(packet)))
            return
        if DEBUG_WRAPPER:
            log("calling %s.%s%s", wo, attr, str(tuple(packet[1:]))[:128])
        GLib.idle_add(method, *packet[1:])
        INJECT_FAULT(proto)


def exec_kwargs(**kwargs) -> dict[str, Any]:
    if WIN32:
        from xpra.platform.win32 import REDIRECT_OUTPUT
        if REDIRECT_OUTPUT:
            # stderr is not valid and would give us this error:
            # WindowsError: [Errno 6] The handle is invalid
            if "stderr" not in kwargs:
                kwargs["stderr"] = open(os.devnull, 'w')
        if not WIN32_SHOWWINDOW:
            startupinfo = subprocess.STARTUPINFO()  # @UndefinedVariable
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW  # @UndefinedVariable
            startupinfo.wShowWindow = 0  # aka win32.con.SW_HIDE
            kwargs["startupinfo"] = startupinfo
    if "stderr" not in kwargs and sys.stderr:
        kwargs["stderr"] = sys.stderr.fileno()
    return kwargs


def exec_env() -> dict[str, str]:
    env = get_exec_env()
    env["XPRA_SKIP_UI"] = "1"
    env["XPRA_FORCE_COLOR_LOG"] = "1"
    return env


class SubprocessCaller:
    """
    This is the caller side, wrapping the subprocess.
    You can call send() to pass packets to it
     which will get converted to method calls on the receiving end,
    You can register for signals, in which case your callbacks will be called
     when those signals are forwarded back.
    (there is no validation of which signals are valid or not)
    """

    def __init__(self, description="wrapper"):
        self.process = None
        self.protocol = None
        self.command = None
        self.description = description
        self.send_queue: SimpleQueue[PacketType] = SimpleQueue()
        self.signal_callbacks: dict[str, list[tuple[Callable, list[Any]]]] = {}
        self.large_packets = []
        # hook a default packet handlers:
        self.connect(CONNECTION_LOST, self.connection_lost)
        self.connect(GIBBERISH, self.gibberish)

    def connect(self, signal: str, cb: Callable, *args) -> None:
        """ gobject style signal registration """
        self.signal_callbacks.setdefault(signal, []).append((cb, list(args)))

    def subprocess_exit(self, *args) -> None:
        # beware: this may fire more than once!
        log("subprocess_exit%s command=%s", args, self.command)
        self._fire_callback("exit")

    def start(self) -> None:
        assert self.process is None, "already started"
        self.process = self.exec_subprocess()
        self.protocol = self.make_protocol()
        self.protocol.start()

    def abort_test(self, action) -> None:
        p = self.process
        if p is None or p.poll():
            raise ConnectionClosedException("cannot %s: subprocess has terminated" % action) from None

    def make_protocol(self) -> SocketProtocol:
        # make a connection using the process stdin / stdout
        conn = TwoFileConnection(self.process.stdin, self.process.stdout,
                                 abort_test=self.abort_test, target=self.description,
                                 socktype=self.description, close_cb=self.subprocess_exit)
        conn.timeout = 0
        protocol = SocketProtocol(conn, self.process_packet, get_packet_cb=self.get_packet)
        setup_fastencoder_nocompression(protocol)
        protocol.large_packets = self.large_packets
        return protocol

    def exec_subprocess(self) -> subprocess.Popen:
        kwargs = exec_kwargs()
        env = self.get_env()
        log("exec_subprocess() command=%s, env=%s, kwargs=%s", self.command, env, kwargs)
        proc = subprocess.Popen(self.command, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                                env=env, start_new_session=True, **kwargs)
        getChildReaper().add_process(proc, self.description, self.command, True, True, callback=self.subprocess_exit)
        return proc

    def get_env(self) -> dict[str, str]:
        env = exec_env()
        env["XPRA_LOG_PREFIX"] = "%s " % self.description
        env["XPRA_FIX_UNICODE_OUT"] = "0"
        # MS Windows wrappers should never wait for input:
        env["XPRA_WAIT_FOR_INPUT"] = "0"
        return env

    def cleanup(self) -> None:
        self.stop()

    def stop(self) -> None:
        self.stop_process()
        self.stop_protocol()

    def stop_process(self) -> None:
        log("%s.stop_process() sending stop request to %s", self, self.description)
        proc = self.process
        if not proc:
            return
        self.process = None
        if proc.poll() is None:
            try:
                proc.terminate()
            except Exception as e:
                log("stop_process() proc=%s", proc, exc_info=True)
                log.warn("Warning: failed to stop the wrapped subprocess %s: %s", proc, e)

    def stop_protocol(self) -> None:
        p = self.protocol
        if p:
            self.protocol = None
            log("%s.stop_protocol() calling %s", self, p.close)
            try:
                p.close()
            except Exception as e:
                log("stop_protocol() protocol=%s", p, exc_info=True)
                log.warn("Warning: failed to close the subprocess connection %s: %s", p, e)

    def connection_lost(self, *args) -> None:
        log("connection_lost%s", args)
        self.stop()

    def gibberish(self, *args) -> None:
        log.warn("%s stopping on gibberish:", self.description)
        log.warn(" %s", repr_ellipsized(args[1], limit=80))
        self.stop()

    def get_packet(self) -> tuple[PacketType, bool, bool]:
        try:
            item = self.send_queue.get(False)
            more = self.send_queue.qsize() > 0
        except Empty:
            item = None
            more = False
        return item, False, more

    def send(self, packet_type: str, *packet_data: PacketElement) -> None:
        packet = (packet_type, *packet_data)
        self.send_queue.put(packet)
        p = self.protocol
        if p:
            p.source_has_more()
            if FLUSH:
                conn = p._conn
                if conn and conn.is_active():
                    conn.flush()
        INJECT_FAULT(p)

    def process_packet(self, proto, packet: PacketType) -> None:
        if DEBUG_WRAPPER:
            log("process_packet(%s, %s)", proto, [str(x)[:32] for x in packet])
        signal_name = str(packet[0])
        self._fire_callback(signal_name, packet[1:])
        INJECT_FAULT(proto)

    def _fire_callback(self, signal_name: str, extra_args=()) -> None:
        callbacks = self.signal_callbacks.get(signal_name, [])
        log("firing callback for '%s': %s", signal_name, callbacks)
        for cb, args in callbacks:
            with log.trap_error(f"Error processing callback {cb} for {signal_name} packet"):
                all_args = list(args) + list(extra_args)
                GLib.idle_add(cb, self, *all_args)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest
from xpra.os_util import WIN32
from xpra.util.io import pollwait
from unit.server_test_util import ServerTestUtil, log


class ProxyServerTest(ServerTestUtil):

    def test_proxy_start_stop(self):
        display = self.find_free_display()
        log(f"using free display={display}")
        cmd = ["proxy", display, "--no-daemon"]
        cmdstr = " ".join(f"'{c}'" for c in cmd)
        proxy = self.run_xpra(cmd)
        r = pollwait(proxy, 5)
        if r is not None:
            self.show_proc_pipes(proxy)
        assert r is None, f"proxy failed to start with cmd={cmdstr}, exit code={r}"
        displays = self.dotxpra.displays()
        assert display in displays, f"proxy display {display!r} not found in {displays}"
        self.check_stop_server(proxy, "stop", display)

    def stop_server(self, server_proc, subcommand, *connect_args):
        if WIN32:
            super().stop_server(server_proc, subcommand, *connect_args)
            return
        log("stop_server%s", (server_proc, subcommand, connect_args))
        if server_proc.poll() is not None:
            return
        server_proc.terminate()
        assert pollwait(server_proc) is not None, f"server process {server_proc} failed to exit"


def main():
    unittest.main()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import gi

gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk  #pylint: disable=wrong-import-position @UnresolvedImport

width = 400
height = 200


def main():
    window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
    #window = gtk.Window(gtk.WINDOW_POPUP)
    window.set_size_request(width, height)
    window.connect("delete_event", Gtk.main_quit)
    vbox = Gtk.VBox()

    def send_net_showing_desktop(v):
        from xpra.x11.gtk.display_source import init_gdk_display_source
        from xpra.x11.bindings.window import constants, X11WindowBindings  #@UnresolvedImport
        init_gdk_display_source()
        X11Window = X11WindowBindings()
        root = window.get_window().get_screen().get_root_window()
        root_xid = root.get_xid()
        SubstructureNotifyMask = constants["SubstructureNotifyMask"]
        SubstructureRedirectMask = constants["SubstructureRedirectMask"]
        event_mask = SubstructureNotifyMask | SubstructureRedirectMask
        X11Window.sendClientMessage(root_xid, root_xid, False, event_mask, "_NET_SHOWING_DESKTOP", v)

    b = Gtk.Button(label="Show Desktop")

    def show_desktop(*_args):
        send_net_showing_desktop(1)

    b.connect('clicked', show_desktop)
    vbox.add(b)

    b = Gtk.Button(label="Not Show Desktop")

    def not_show_desktop(*_args):
        send_net_showing_desktop(0)

    b.connect('clicked', not_show_desktop)
    vbox.add(b)

    window.add(vbox)
    window.show_all()
    Gtk.main()
    return 0


if __name__ == "__main__":
    main()
